import fnmatch
import json
import logging
import os
import subprocess
import sys
import threading
from collections import OrderedDict
from functools import partial
from pathlib import Path
from typing import Callable, List, Tuple
from warnings import warn

import pyqtgraph as pg
from qtpy import API_NAME, QtCore, QtWidgets, uic


class OrderedAttrDict(object):

    def __init__(self):
        self._odict = OrderedDict()

    def add(self, name, obj):
        self._odict[name] = obj
        self.__dict__[name] = obj
        return obj

    def keys(self):
        return self._odict.keys()

    def values(self):
        return self._odict.values()

    def items(self):
        return self._odict.items()

    def __len__(self):
        return len(self._odict)

    def __getitem__(self, key):
        return self._odict[key]

    def __contains__(self, k):
        return self._odict.__contains__(k)


def open_file(filepath):
    """
    Cross-platform file opener since a native module does not yet exist per
    https://stackoverflow.com/questions/17317219/is-there-an-platform-independent-equivalent-of-os-startfile
    """
    try:
        if sys.platform.startswith("win"):  # Windows
            os.startfile(filepath)
        elif sys.platform.startswith("darwin"):  # macOS
            subprocess.call(["open", filepath])
        else:  # linux
            subprocess.call(["xdg-open", filepath])
    except Exception as e:
        print(f"Error opening file {filepath}: {e}")


def sibling_path(a, b):
    """
    Returns the path of a filename *b* in the same folder as *a*
    (i.e. dirname(a)/b )

    ==============  ==============================
    **Arguments:**
    *a*             directory name of pathname *a*
    b               path suffix
    ==============  ==============================

    :returns: combined path of above arguments.
    """
    return os.path.join(os.path.dirname(a), b)


def load_qt_ui_file(ui_filename):
    """
    Loads a QT user interface file (files ending in .ui).
    This function is typically called from :class:`Measurement` level modules.
    """
    ### PySide version
    # ui_loader = QtUiTools.QUiLoader()
    # ui_file = QtCore.QFile(ui_filename)
    # ui_file.open(QtCore.QFile.ReadOnly)
    # ui = ui_loader.load(ui_file)
    # ui_file.close()
    ### qtpy / PyQt version
    ui = uic.loadUi(ui_filename)
    return ui


def load_qt_ui_from_pkg(package, filename):
    if API_NAME == "PySide6":
        warn(
            f"App runs normally, only packaging DataBrowser not support with {API_NAME}. Use PyQt6 instead.",
            RuntimeWarning,
        )
        return load_qt_ui_file(Path(__file__).parent / "data_browser" / filename)
    import pkgutil
    from io import StringIO

    ui_data = pkgutil.get_data(package, filename).decode()
    ui_io = StringIO(ui_data)
    ui = uic.loadUi(ui_io)
    return ui


def confirm_on_close(
    widget,
    title="Close ScopeFoundry?",
    message="Do you wish to shut down ScopeFoundry?",
    func_on_close=None,
):
    """
    Calls the :class:`ConfirmCloseEventEater` class which asks for user
    confirmation in a pop-up dialog upon closing the ScopeFoundry app.
    """
    widget.closeEventEater = ConfirmCloseEventEater(title, message, func_on_close)
    widget.installEventFilter(widget.closeEventEater)


class ConfirmCloseEventEater(QtCore.QObject):
    """
    Tells the Qt to confirm the closing of the app by the user with a pop-up
    confirmation dialog.
    """

    def __init__(
        self,
        title="Close ScopeFoundry?",
        message="Do you wish to shut down ScopeFoundry?",
        func_on_close=None,
    ):
        QtCore.QObject.__init__(self)
        self.title = title
        self.message = message
        self.func_on_close = func_on_close

    def eventFilter(self, obj, event):
        """
        Listens for QtCore.QEvent.Close signal and asks the user whether to
        close the app in a pop-up dialog."""
        if event.type() == QtCore.QEvent.Close:
            # eat close event
            logging.debug("close")
            reply = QtWidgets.QMessageBox.question(
                None,
                self.title,
                self.message,
                QtWidgets.QMessageBox.Yes,
                QtWidgets.QMessageBox.No,
            )
            if reply == QtWidgets.QMessageBox.Yes:
                logging.debug("closing")
                if self.func_on_close:
                    self.func_on_close()
                QtWidgets.QApplication.quit()
                event.accept()
            else:
                event.ignore()
            return True
        else:
            # standard event processing
            return QtCore.QObject.eventFilter(self, obj, event)


def ignore_on_close(widget):
    """
    Calls the :class:`IgnoreCloseEventEater` class which intercepts the
    QtCore.QEvent.Close signal and prevents the deletion of subwindows and their
    associated objects.
    """
    widget.ignoreCloseEventEater = IgnoreCloseEventEater()
    widget.installEventFilter(widget.ignoreCloseEventEater)


class IgnoreCloseEventEater(QtCore.QObject):
    """
    This class is utilized to prevent the closing of important subwindows
    and prevent the deletion of related python objects. Tells the Qt event
    manager to ignore calls to close the subwindow when user hits [x] on
    the subwindow.
    """

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Close:
            # eat close event
            event.ignore()
            return True
        else:
            # standard event processing
            return QtCore.QObject.eventFilter(self, obj, event)


def replace_widget_in_layout(
    old_widget, new_widget, retain_font=True, retain_sizePolicy=True, retain_size=True
):
    """
    Replace a widget with a new widget instance in an existing layout.

    returns new_widget

    Useful for replacing placeholder widgets created with QtCreator
    with custom widgets
    Note: currently works for widgets in  QGridLayouts only

    inspiration from:
    -- http://stackoverflow.com/questions/4625102/how-to-replace-a-widget-with-another-using-qt
    -- http://stackoverflow.com/questions/24189202/how-to-get-the-row-column-location-of-a-widget-in-a-qgridlayout
    """

    # find position of old_widget
    layout = old_widget.parentWidget().layout()
    index = layout.indexOf(old_widget)
    row, column, rowSpan, colSpan = layout.getItemPosition(index)

    # transfer retained attributes of old_widget to  new_widget
    if retain_font:
        new_widget.setFont(old_widget.font())
    if retain_sizePolicy:
        new_widget.setSizePolicy(old_widget.sizePolicy())
    if retain_size:
        new_widget.setMinimumSize(old_widget.minimumSize())
        new_widget.setMaximumSize(old_widget.maximumSize())

    # remove old widget
    layout.removeWidget(old_widget)
    old_widget.close()

    # add new widget
    layout.addWidget(new_widget, row, column, rowSpan, colSpan)
    layout.update()

    return new_widget


def auto_connect_widget_in_ui(ui, lq):
    """
    Automatically finds widgets in a User Interface QWidget *ui*
    created from *load_qt_ui_file*
    that match the name of the LoggedQuantity *lq* and makes
    a bidirectional connection
    """
    for widget_name, widget in ui.__dict__.items():
        if lq.name in widget_name:
            lq.connect_to_widget(widget)


def replace_spinbox_in_layout(old_widget, **kwargs):
    new_spinbox = pg.SpinBox()
    return replace_widget_in_layout(old_widget, new_widget=new_spinbox, **kwargs)


def groupbox_show_contents(groupbox, show=True):
    layout = groupbox.layout()

    for i in range(layout.count()):
        layout.itemAt(i).widget().setVisible(show)


def print_all_connected(qobject, signal=None):
    if signal is None:
        signals = qobject.signals()
    else:
        signals = [signal]
    for signal in qobject.signals():
        for slot in qobject.connectedSlots():
            print(slot)


def print_signals_and_slots(obj):
    # http://visitusers.org/index.php?title=PySide_Recipes
    for i in range(obj.metaObject().methodCount()):
        m = obj.metaObject().method(i)
        if m.methodType() == QtCore.QMetaMethod.MethodType.Signal:
            print(
                "SIGNAL: sig=",
                m.signature(),
                "hooked to nslots=",
                obj.receivers(QtCore.SIGNAL(m.signature())),
            )
        elif m.methodType() == QtCore.QMetaMethod.MethodType.Slot:
            print("SLOT: sig=", m.signature())


def get_logger_from_class(obj):
    """returns a named Logger from the logging package using the
    full name of the class of the object (obj) as the log name
    """
    # return logging.getLogger(f"{obj.__module__}.{obj.__class__.__name__})
    return logging.getLogger(obj.__class__.__name__)


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def bool2str(v):
    return {False: "False", True: "True"}[v]


class DummyLock:
    def acquire(self): ...

    def release(self): ...

    def __enter__(self):
        return self

    def __exit__(self, *args): ...


def QLock(mode: int = 0):
    """this function was introduced for backward compatibility as PyQt6 initializer no longer accepts a mode keyword,
    in the future used either QNonReEntrantLock or QReEntrantLock"""

    qt_version = os.environ["QT_API"].lower()[-1]
    # print('detected qt_version', qt_version)
    if qt_version in ("4", "5"):

        class Q45Lock(QtCore.QMutex):
            """used if qt 4 or 5, mainly for backwards compatibility"""

            def acquire(self):
                self.lock()

            def release(self):
                self.unlock()

            def __enter__(self):
                self.acquire()
                return self

            def __exit__(self, *args):
                self.release()

        return Q45Lock(mode=mode)

    elif qt_version in ("6",):

        class QNonReEntrantLock(QtCore.QMutex):
            def acquire(self):
                self.lock()

            def release(self):
                self.unlock()

            def __enter__(self):
                self.acquire()
                return self

            def __exit__(self, *args):
                self.release()

        class QReEntrantLock(QtCore.QRecursiveMutex):
            def acquire(self):
                self.lock()

            def release(self):
                self.unlock()

            def __enter__(self):
                self.acquire()
                return self

            def __exit__(self, *args):
                self.release()

        if mode == 1:
            return QReEntrantLock()
        else:
            return QNonReEntrantLock()
    return DummyLock()


# https://stackoverflow.com/questions/5327614/logging-lock-acquire-and-release-calls-in-multi-threaded-application
class LogLock(object):
    def __init__(self, name):
        self.name = str(name)
        self.lock = threading.Lock()
        self.log = logging.getLogger("LogLock")

    def acquire(self, blocking=True):
        self.log.debug("{0:x} Trying to acquire {1} lock".format(id(self), self.name))
        ret = self.lock.acquire(blocking)
        if ret == True:
            self.log.debug("{0:x} Acquired {1} lock".format(id(self), self.name))
        else:
            self.log.debug(
                "{0:x} Non-blocking aquire of {1} lock failed".format(
                    id(self), self.name
                )
            )
        return ret

    def release(self):
        self.log.debug("{0:x} Releasing {1} lock".format(id(self), self.name))
        self.lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()
        return False  # Do not swallow exceptions


def find_matches(keys, patterns):
    "returns the keys that fit any of the pattern"
    return [key for key in keys if any(fnmatch.fnmatch(key, pat) for pat in patterns)]


def filter_with_patterns(
    keys, include_patterns=None, exclude_patterns=None
):  # -> list[str]:
    """returns keys to match include and exclude patterns"""
    if include_patterns:
        # reduce possible keys to the ones satisfying the pattern
        keys = find_matches(keys, include_patterns)
    if exclude_patterns:
        keys_to_exclude = find_matches(keys, exclude_patterns)
        keys = [key for key in keys if key not in keys_to_exclude]
    return keys


def get_scopefoundry_version():
    try:
        # only works if installed with pip (hence you are not a developer)
        # https://stackoverflow.com/questions/67085041/how-to-specify-version-in-only-one-place-when-using-pyproject-toml
        from importlib import metadata

        return metadata.version("ScopeFoundry")
    except metadata.PackageNotFoundError:
        pass

    return "dev 1.6+"


def itemize_launchers(
    docs_path: Path, launch_browser: Callable, include_parents_readme: bool = True
) -> List[Tuple[str, Callable]]:

    if not docs_path:
        return []

    items = [("...", partial(open_file, str(docs_path)))]

    if include_parents_readme:
        for path in docs_path.parent.iterdir():
            if path.name.lower() == "readme.md":
                items.append((path.name, partial(open_file, str(path))))

    if not docs_path.exists():
        return items

    for path in docs_path.iterdir():
        if path.name == "links.json":
            items += get_links(path, launch_browser)
        elif path.is_file() and path.name.startswith("__"):
            continue
        elif path.is_file():
            items.append((path.name, partial(open_file, str(path))))

    return items


def get_links(fname, launch_browser):
    with open(fname, "r") as file:
        return [
            (name, partial(launch_browser, url))
            for name, url in json.load(file).items()
        ]


def get_child_path(parent_class) -> Path:
    """folder where a subclass is stored"""
    module = sys.modules[parent_class.__class__.__module__]
    if module.__name__ == "__main__":
        return Path.cwd()
    elif hasattr(module, "__file__"):
        return Path(module.__file__).parent


def init_docs_path(docs_path, settings):
    if not docs_path.exists():
        docs_path.mkdir()
        with open(docs_path / "links.json", "w") as f:
            json.dump({}, f)

    with open(docs_path / f"__place_documentation_here_readme_first.md", "w") as f:
        f.write(f"**this file will be overwriten! **\n\n")
        f.write(
            " - Write documentation describing your app, files starting __ are ignored.\n\n"
        )
        f.write(f" - or add to `links.json`\n\n")

        f.write(f"## Settings\n\n")
        for name in settings.keys():
            lq = settings.get_lq(name)
            f.write(f" - *{lq.name}*: {lq.description}\n\n")



from ScopeFoundry import HardwareComponent
from ScopeFoundry.examples.ScopeFoundryHW.simulon_xyz_stage.simulon_xyz_stage_dev import (
    SimulonXYZStageDev,
)


class SimulonXYZStageHW(HardwareComponent):

    name = "simulon_xyz_stage"

    def setup(self):
        position_params = {
            "dtype": float,
            "initial": -1,
            "vmin": -1,
            "vmax": 100,
            "si": False,
            "unit": "um",
            "spinbox_decimals": 3,
        }

        self.x_position = self.settings.New("x_position", ro=True, **position_params)
        self.y_position = self.settings.New("y_position", ro=True, **position_params)
        self.z_position = self.settings.New("z_position", ro=True, **position_params)
        self.x_target_position = self.settings.New(
            "x_target_position", ro=False, **position_params
        )
        self.y_target_position = self.settings.New(
            "y_target_position", ro=False, **position_params
        )
        self.z_target_position = self.settings.New(
            "z_target_position", ro=False, **position_params
        )

    def connect(self):

        dev = self.stage_device = SimulonXYZStageDev(debug=self.debug_mode.val)

        self.x_position.connect_to_hardware(read_func=dev.read_x)
        self.y_position.connect_to_hardware(read_func=dev.read_y)
        self.z_position.connect_to_hardware(read_func=dev.read_z)

        self.x_target_position.connect_to_hardware(write_func=dev.write_x)
        self.y_target_position.connect_to_hardware(write_func=dev.write_y)
        self.z_target_position.connect_to_hardware(write_func=dev.write_z)

    def disconnect(self):

        self.settings.disconnect_all_from_hardware()

        if hasattr(self, "stage_device"):
            self.stage_device.close()
            del self.stage_device



"""
Created on Sep 17, 2021

@author: Benedikt Ursprung
"""

import json
import os
from pathlib import Path
import time
from builtins import getattr
from typing import List, Tuple, Dict, Union

from qtpy import QtWidgets, QtCore


from ScopeFoundry import Measurement
from .item_types.editor_base_controller import EditorBaseController
from .item_types.editor_base_ui import EditorBaseUI
from .item_types.item_list import ItemList
from .item_types import (
    BaseItem,
    ExecFunctionEditorUI,
    InterationsEditorController,
    IterationsEditorUI,
    IterruptIfEditorUI,
    NewDirEditorUI,
    PauseEditorUI,
    ReadFromHardWareEditorUI,
    RunMeasurementEditorUI,
    SaveDirToParentEditorUI,
    TimeoutEditorUI,
    UpdateSettingEditorUI,
    WaitUntilEditorUI,
    link_iteration_items,
    new_item,
)


class Sequencer(Measurement):
    name = "sequencer"

    def setup(self):

        self.item_list = ItemList()

        self.settings.New(
            "cycles",
            int,
            initial=1,
            description="number of times the sequence is executed",
        )
        self.settings.New("paused", bool, initial=False)
        self.recipe_folder = self.settings.new_file(
            "recipe_folder", initial=str(Path.cwd()), is_dir=True
        )
        self.seq_file = self.settings.new_file(
            "sequence_file",
            initial="",
            is_dir=False,
            file_filters=("Sequence (*.json)"),
        )

        self.iter_values = {}
        self.letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.current_item = None

    def setup_figure(self):
        # measurement controls and settings
        meas_layout = QtWidgets.QHBoxLayout()
        meas_layout.addWidget(self.settings.get_lq("cycles").new_default_widget())
        meas_layout.addWidget(self.new_start_stop_button())
        btn = self.settings.get_lq("paused").new_pushButton(
            texts=["pause", "resume"], colors=[None, "rgba( 0, 255, 0, 220)"]
        )
        meas_layout.addWidget(btn)

        # select file combobox
        self.load_file_comboBox = QtWidgets.QComboBox()
        self.load_file_comboBox.setToolTip("previously saved sequences")
        self.update_load_file_comboBox()
        self.load_file_comboBox.currentTextChanged.connect(
            self.on_load_file_comboBox_changed
        )

        # item list
        item_list_scroll_area = QtWidgets.QScrollArea()
        # item_list_scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        # item_list_scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        item_list_scroll_area.setWidgetResizable(True)
        item_list_scroll_area.setWidget(self.item_list.get_view())
        self.item_list.connect_item_double_clicked(self.item_double_clicked)

        # controls
        file_layout = QtWidgets.QHBoxLayout()
        self.remove_pushButton = QtWidgets.QPushButton("remove selected item")
        self.remove_pushButton.setToolTip("DEL")
        self.remove_pushButton.clicked.connect(self.on_remove_item)
        file_layout.addWidget(self.remove_pushButton)
        btn = QtWidgets.QPushButton("save ...")
        btn.clicked.connect(self.on_save)
        file_layout.addWidget(btn)
        btn = QtWidgets.QPushButton("load ...")
        btn.clicked.connect(self.seq_file.file_browser)
        file_layout.addWidget(btn)
        btn = QtWidgets.QPushButton("run selected item")
        btn.setToolTip("SPACEBAR")
        btn.clicked.connect(self.on_run_item_and_proceed)
        file_layout.addWidget(btn)
        self.show_editor_checkBox = QtWidgets.QCheckBox("show|hide editor")
        file_layout.addWidget(self.show_editor_checkBox)

        # Combine so far
        top_widget = QtWidgets.QWidget()
        self.layout = top_layout = QtWidgets.QVBoxLayout(top_widget)
        top_layout.addLayout(meas_layout)
        top_layout.addWidget(self.load_file_comboBox)
        top_layout.addWidget(item_list_scroll_area)
        top_layout.addLayout(file_layout)

        # Editors
        self.editor_widget = QtWidgets.QWidget()
        self.editor_layout = QtWidgets.QVBoxLayout(self.editor_widget)
        editor_scroll_area = QtWidgets.QScrollArea()
        # editor_scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        # editor_scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        editor_scroll_area.setWidgetResizable(True)
        editor_scroll_area.setWidget(self.editor_widget)
        self.show_editor_checkBox.stateChanged.connect(editor_scroll_area.setVisible)
        self.show_editor_checkBox.setChecked(True)

        paths = self.app.get_setting_paths()
        funcs = get_all_functions(self.app)
        self.editors: Dict[str, EditorBaseController] = {}
        self.add_editor(ReadFromHardWareEditorUI(self, paths))
        self.add_editor(UpdateSettingEditorUI(self, paths))
        self.add_editor(RunMeasurementEditorUI(self))
        self.add_editor(WaitUntilEditorUI(self, paths))
        self.add_editor(WaitUntilEditorUI(self, paths))
        self.add_editor(ExecFunctionEditorUI(self, funcs))
        self.add_editor(PauseEditorUI(self))
        self.add_editor(IterruptIfEditorUI(self, paths))
        self.add_editor(NewDirEditorUI(self))
        self.add_editor(SaveDirToParentEditorUI(self))
        self.add_editor(TimeoutEditorUI(self))
        self.add_iteration_editor(IterationsEditorUI(self, paths))

        for editor in self.editors.values():
            self.editor_layout.addWidget(editor.ui.group_box)

        self.editor_widget.keyPressEvent = self._editorKeyPressEvent
        self.item_list.get_view().keyReleaseEvent = self._keyReleaseEvent

        # connect logged quantities
        self.recipe_folder.add_listener(self.update_load_file_comboBox)
        self.seq_file.add_listener(self.on_load)

        # try to load a file
        text = self.load_file_comboBox.currentText()
        if text in self.seq_fnames:
            self.seq_file.update_value(self.seq_fnames[text])

        self.ui = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal)
        self.ui.addWidget(top_widget)
        self.ui.addWidget(editor_scroll_area)
        self.ui.setStretchFactor(0, 1)
        self.ui.setStretchFactor(1, 1)

    def add_editor(self, editor_ui: EditorBaseUI):
        self.editors[editor_ui.item_type] = EditorBaseController(editor_ui, self)

    def add_iteration_editor(self, editor_ui: IterationsEditorUI):
        editor = InterationsEditorController(editor_ui, self)
        self.editors["start-iteration"] = editor
        self.editors["end-iteration"] = editor

    def _editorKeyPressEvent(self, event):
        if not event.modifiers() & QtCore.Qt.ControlModifier:
            return
        if not event.key() in (QtCore.Qt.Key_R, QtCore.Qt.Key_N):
            return
        fw = self.editor_widget.focusWidget()
        # find editor with focused widget
        for e in self.editors.values():
            gb = e.ui.group_box
            if fw in gb.findChildren(type(fw), fw.objectName()):
                if event.key() == QtCore.Qt.Key_R:
                    e.on_replace_func()
                if event.key() == QtCore.Qt.Key_N:
                    e.on_new_func()

    def _keyReleaseEvent(self, event):
        if event.key() == QtCore.Qt.Key_Delete:
            self.item_list.remove()
        if event.key() == QtCore.Qt.Key_Space:
            self.on_run_item_and_proceed()
        if event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            item = self.item_list.get_current_item()
            self.item_double_clicked(item)

    def is_valid_file(self, fname: Path) -> bool:
        if fname.name == ("links.json"):
            return False
        with open(fname, "r") as f:
            items = json.loads(f.read())
        for kwargs in items:
            if not isinstance(kwargs, dict) or not "type" in kwargs:
                return False
        return True

    def update_load_file_comboBox(self):
        root = Path(self.settings["recipe_folder"])
        fnames = [fname for fname in root.rglob("*.json") if self.is_valid_file(fname)]
        index0 = self.load_file_comboBox.currentIndex()
        self.load_file_comboBox.clear()
        self.seq_fnames = {}
        for fname in fnames:
            abbrev_fname = fname.relative_to(root)
            self.seq_fnames.update({str(abbrev_fname): fname})
        self.load_file_comboBox.addItems(list(self.seq_fnames.keys()))
        index = max(0, min(index0, self.load_file_comboBox.count() - 1))
        self.load_file_comboBox.setCurrentIndex(index)

    def insert_load_files(self, fname: Path):
        root = Path(self.settings["recipe_folder"])
        abbrev_fname = str(fname.relative_to(root))
        if abbrev_fname in self.seq_fnames:
            return
        self.seq_fnames.update({abbrev_fname: fname})
        self.load_file_comboBox.insertItem(0, abbrev_fname)
        self.load_file_comboBox.setCurrentIndex(0)

    def on_load_file_comboBox_changed(self, abbrev_fname):
        self.load_file(self.seq_fnames[abbrev_fname])

    def next_iter_id(self) -> str:
        return self.letters[self.item_list.count_type("start-iteration")]

    def on_remove_item(self):
        self.item_list.remove(None)

    def on_save(self) -> str:
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, caption="Save Sequence", filter="Sequence (*.json)"
        )
        if fname:
            self.save_to_file(fname)
        self.update_load_file_comboBox()
        return fname

    def save_to_file(self, fname):
        with open(fname, "w+") as f:
            f.write(json.dumps(self.item_list.as_dicts(), indent=1))

    def on_load(self) -> str:
        fname = Path(self.seq_file.val)
        if fname.exists() and not fname.is_dir():
            self.load_file(fname)
            self.insert_load_files(fname)
        return fname

    def load_file(self, fname):
        self.item_list.clear()
        with open(fname, "r") as f:
            lines = json.loads(f.read())
        for kwargs in lines:
            item_type = kwargs.pop("type")
            item = new_item(self, item_type, **kwargs)
            self.item_list.add(item)
        success = link_iteration_items(self.item_list)
        if not success:
            print("invalid list")

    def on_run_item(self) -> Tuple[BaseItem, Union[None, BaseItem]]:
        item = self.item_list.get_current_item()
        if item.item_type == "measurement":
            print("WARNING running individually measurement not supported")
            return (item, None)
        return (item, self.item_list.get_current_item().visit())

    def on_run_item_and_proceed(self):
        item, next_item = self.on_run_item()
        if next_item is None:
            row = self.item_list.get_row(item)
            next_item = self.item_list.get_item(row + 1)
        self.item_list.set_current_item(next_item)

    def item_double_clicked(self, item: BaseItem):
        self.editors[item.item_type].ui.set_kwargs(**item.kwargs)

    def run(self):
        success = link_iteration_items(self.item_list)
        if not success:
            print("invalid list")

        N = self.item_list.count()
        for i in range(N):
            self.item_list.get_item(i).reset()

        for q in range(self.settings["cycles"]):
            # pct = int(100 * q / self.settings['cycles'])
            # self.set_progress(pct)
            if self.interrupt_measurement_called:
                break

            # go through list
            row = 0
            while row < self.item_list.count():
                while self.settings["paused"]:
                    if self.interrupt_measurement_called:
                        break
                    time.sleep(0.03)

                self.current_item = item = self.item_list.get_item(row)

                resp = item.visit()
                if resp is None:
                    # go to next item
                    row += 1
                else:
                    # jump to item returned
                    row = self.item_list.get_row(resp)

                if self.interrupt_measurement_called:
                    break

        self.current_item = None

    def update_display(self):
        for i in range(self.item_list.count()):
            item = self.item_list.get_item(i)
            item.setSelected(item == self.current_item)

    def shutdown(self):
        os.system("shutdown /s /f /t 1")

    def restart(self):
        os.system("restart /r /f /t 1")


def get_all_functions(app) -> List[str]:
    funcs = []

    def append_objs_callables(obj, from_app_path):
        for a in dir(obj):

            # exclude deprecated functions
            if isinstance(obj, Measurement) and a in ("gui",):
                continue

            try:  # Not sure why some python version seem to need this block
                if callable(getattr(obj, a)) and not a.startswith("__"):
                    funcs.append(f"{from_app_path}{obj.name}.{a}")
            except AttributeError as e:
                print(e)

    append_objs_callables(app, "")
    for m in app.measurements.values():
        append_objs_callables(m, "measurements.")
    for h in app.hardware.values():
        append_objs_callables(h, "hardware.")
    return funcs



from enum import Enum
from functools import partial
import json
from pathlib import Path
from typing import Dict

from ScopeFoundry.tools.features.utils import (
    mk_authors,
    mk_dates,
    to_class_name,
    write_templated_file,
)


class ComTypes(Enum):
    OTHER = "OTHER"
    DLL = "DLL"
    SERIAL = "SERIAL"
    # VISA = 'VISA'


def new_hw(
    company: str = "thorlabs",
    model: str = "XX6000_glavo_mirror",
    authors: str = "Benedikt Ursprung, Mark E.Ziffer",
    pretty_name: str = "galvo_mirror",
    overwrite_existing_module: bool = True,
    com_type: str = ComTypes.DLL.value,
) -> Dict[str, str]:
    infos = gather_infos(
        company, model, authors, pretty_name, overwrite_existing_module
    )

    path_to_module = Path(infos["PATH_TO_MODULE"])
    path_to_module.mkdir(exist_ok=True)
    root = Path(__file__).parent.parent

    write = partial(
        write_templated_file,
        substitutions=infos,
        path_to_target=path_to_module,
    )

    write(root / "templates/_readout.py", infos["READOUT_FILE_NAME"])
    write(root / "templates/_test_app.py", infos["TEST_APP_FILE_NAME"])
    write(root / "templates/___init__.py", "__init__.py")
    write(root / "templates/_LICENSE", "LICENSE")
    write(root / "templates/_README.md", "README.md")
    write(root / "templates/.gitignore", ".gitignore")

    if com_type == ComTypes.DLL.value:
        write(root / "templates/_dll_dev.py", infos["DEV_FILE_NAME"])
        write(root / "templates/_dll_hw.py", infos["HW_FILE_NAME"])
    if com_type == ComTypes.SERIAL.value:
        write(root / "templates/_serial_dev.py", infos["DEV_FILE_NAME"])
        write(root / "templates/_serial_hw.py", infos["HW_FILE_NAME"])
    if com_type == ComTypes.OTHER.value:
        write(root / "templates/_serial_dev.py", infos["DEV_FILE_NAME"])
        write(root / "templates/_serial_hw.py", infos["HW_FILE_NAME"])

    docs_path = path_to_module / "docs"
    docs_path.mkdir()

    with open(docs_path / "links.json", "x", encoding="utf-8") as file:
        content = {
            "What's this?": "https://scopefoundry.org/docs/30_tips-and-tricks/document-your-components/"
        }
        json.dump(content, file)
    return infos


def gather_infos(
    company: str,
    model: str,
    authors: str,
    pretty_name: str,
    overwrite_existing_module: bool,
) -> Dict[str, str]:
    module_name = f"{company.lower()}_{model.lower()}"

    path_to_module = assert_ScopeFoundryHW_directory() / module_name

    if path_to_module.exists() and not overwrite_existing_module:
        raise FileExistsError(
            f"{module_name} already exists! You can use overwrite flag at your own risk"
        )

    if pretty_name is None or pretty_name == "":
        pretty_name = model.lower()
    pretty_name = pretty_name.lower()

    dev_file_name = f"{pretty_name}_dev.py"
    dev_class_name = to_class_name(f"{model.upper()}_DEV")
    import_dev = f"from .{dev_file_name.rstrip('.py')} import {dev_class_name}"

    hw_file_name = f"{pretty_name.lower()}_hw.py"
    hw_class_name = f"{to_class_name(model.upper())}HW"

    readout_file_name = f"{pretty_name.lower()}_readout.py"
    readout_class_name = to_class_name(f"{model.upper()}_Readout")

    test_app_file_name = f"{ pretty_name.lower()}_test_app.py"

    init_imports = f"from .{hw_file_name.rstrip('.py')} import {hw_class_name} \nfrom .{readout_file_name.rstrip('.py')} import {readout_class_name}"
    test_app_imports = f"from ScopeFoundryHW.{module_name.rstrip('.py')} import {hw_class_name}, {readout_class_name}"

    add_to_app = f"""
        {test_app_imports}
        self.add_hardware({hw_class_name}(self))
        self.add_measurement({readout_class_name}(self))
"""

    infos = {
        "DEV_FILE_NAME": dev_file_name,
        "IMPORT_DEV": import_dev,
        "DEV_CLASS_NAME": dev_class_name,
        "HW_FILE_NAME": hw_file_name,
        "HW_CLASS_NAME": hw_class_name,
        "READOUT_CLASS_NAME": readout_class_name,
        "READOUT_NAME": f"{pretty_name.lower()}_readout",
        "READOUT_FILE_NAME": readout_file_name,
        "TEST_APP_NAME": f"{pretty_name.lower()}_test_app",
        "HW_NAME": pretty_name.lower(),
        "INIT_IMPORTS": init_imports,
        "IMPORTS_IN_TEST_APP": test_app_imports,
        "MODULE_NAME": module_name,
        "COMPANY": " ".join(company.split("_")),
        "MODEL": " ".join(model.split("_")),
        "ADD_TO_APP": add_to_app,
        "TEST_APP_FILE_NAME": test_app_file_name,
        "PATH_TO_MODULE": path_to_module,
    }
    infos.update(mk_dates())
    infos.update(mk_authors(authors))
    return infos


def assert_ScopeFoundryHW_directory():
    cwd = Path.cwd()
    if has_ScopeFoundryHW_directory(cwd):
        return cwd / "ScopeFoundryHW"
    elif has_ScopeFoundryHW_directory(cwd.parent):
        return cwd.parent / "ScopeFoundryHW"
    elif has_ScopeFoundryHW_directory(cwd.parent.parent):
        return cwd.parent.parent / "ScopeFoundryHW"
    else:
        try:
            from qtpy import QtWidgets

            a = QtWidgets.QFileDialog.getExistingDirectory(
                caption="Select your ScopeFoundryHW folder or create one",
                directory=str(cwd),
            )
            return Path(a)
        except Exception as e:
            print(
                "No ScopeFoundryHW directory found and no qtpy available: cd to your project folder or pip install qtpy"
            )


def has_ScopeFoundryHW_directory(path: Path) -> bool:
    for p in path.iterdir():
        if p.is_dir() and p.name == "ScopeFoundryHW":
            return True
    return False


if __name__ == "__main__":
    new_hw()



import unittest

from pyqtgraph.widgets.SpinBox import SpinBox
from qtpy import QtWidgets

from ScopeFoundry.base_app.base_app import BaseApp
from ScopeFoundry.widgets import MinMaxQSlider


class LQConnectToWidgetTest(unittest.TestCase):

    def setUp(self):
        self.app = app = BaseApp([])
        self.settings = app.settings
        self.str = self.settings.New("str", str, initial="0")
        self.float = self.settings.New("float", float, vmin=-100, vmax=100)
        self.int = self.settings.New("int", int)
        self.bool = self.settings.New("bool", bool)
        self.choices = self.settings.New(
            "choices", int, choices=(("one", 1), ("two", 2), ("three", 3))
        )
        self._dir = self.settings.New("dir", "file", initial="test_dir", is_dir=True)

    def tearDown(self):
        # self.app.on_close()
        self.app.qtapp.exit()
        del self.app

    def test_qdouble_spinbox(self):
        wf = QtWidgets.QDoubleSpinBox()
        wi = QtWidgets.QDoubleSpinBox()

        wf.setValue(3.0)
        wi.setValue(3)

        # test on connection widget take lq value
        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)
        self.assertEqual(wf.value(), 0.0)
        self.assertEqual(wi.value(), 0)

        # lq to widget
        self.float.update_value(1.0)
        self.int.update_value(1.0)
        self.assertEqual(wf.value(), 1.0)
        self.assertEqual(wi.value(), 1)

        # widget to lq
        wf.setValue(2)
        wi.setValue(2)
        self.assertEqual(self.float.value, 2)
        self.assertEqual(self.int.value, 2)

    def test_pyqtgraph_spinbox(self):
        wf = SpinBox()
        wi = SpinBox()

        wf.setValue(3.0)
        wi.setValue(3)

        # test on connection widget take lq value
        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)
        self.assertEqual(wf.value(), 0.0)
        self.assertEqual(wi.value(), 0)

        # lq to widget
        self.float.update_value(1.0)
        self.int.update_value(1.0)
        self.assertEqual(wf.value(), 1)
        self.assertEqual(wf.value(), 1)

        # widget to lq
        wf.setValue(2)
        wi.setValue(2)
        self.assertEqual(self.float.value, 2)
        self.assertEqual(self.int.value, 2)

    def test_qlabel(self):
        ws = QtWidgets.QLabel()
        wf = QtWidgets.QLabel()
        wi = QtWidgets.QLabel()

        ws.setText("3")
        wf.setText("3")
        wi.setText("3")

        # test on connection widget take lq value
        self.str.connect_to_widget(ws)
        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)
        self.assertEqual(ws.text(), "0")
        self.assertEqual(wf.text(), "0")
        self.assertEqual(wi.text(), "0")

        # lq to widget
        self.str.update_value("1")
        self.float.update_value(1.1)
        self.int.update_value(1)
        self.assertEqual(ws.text(), "1")
        self.assertEqual(wf.text(), "1.1")
        self.assertEqual(wi.text(), "1")

        # widget to lq
        # NA

    def test_qline_edit(self):
        ws = QtWidgets.QLineEdit()
        wf = QtWidgets.QLineEdit()
        wi = QtWidgets.QLineEdit()

        ws.setText("3")
        wf.setText("3")
        wi.setText("3")

        # test on connection widget take lq value
        self.str.connect_to_widget(ws)
        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)
        self.assertEqual(ws.text(), "0")
        self.assertEqual(wf.text(), "0")
        self.assertEqual(wi.text(), "0")

        # lq to widget
        self.str.update_value("1")
        self.float.update_value(1.1)
        self.int.update_value(1)
        self.assertEqual(ws.text(), "1")
        self.assertEqual(wf.text(), "1.1")
        self.assertEqual(wi.text(), "1")

        # widget to lq
        ws.setText("2")
        wf.setText("2")
        wi.setText("2")
        ws.editingFinished.emit()
        wf.editingFinished.emit()
        wi.editingFinished.emit()
        self.assertEqual(self.str.value, "2")
        self.assertEqual(self.float.value, 2.0)
        self.assertEqual(self.int.value, 2)

    def test_q_plain_text_edit(self):

        ws = QtWidgets.QPlainTextEdit()
        ws.document().setPlainText("3")

        # test on connection widget take lq value
        self.str.connect_to_widget(ws)
        self.assertEqual(ws.document().toPlainText(), "0")

        # lq to widget
        self.str.update_value("1")
        self.assertEqual(ws.document().toPlainText(), "1")

        # widget to lq
        ws.document().setPlainText("2")
        self.assertEqual(self.str.value, "2")

    def test_check_box(self):
        wb = QtWidgets.QCheckBox()

        # print("checkbox1")
        wb.setChecked(True)

        # test on connection widget take lq value
        self.bool.connect_to_widget(wb)
        self.assertEqual(wb.isChecked(), False)

        # lq to widget
        # print("checkbox2")

        self.bool.update_value(True)
        self.assertEqual(wb.isChecked(), True)

        # widget to lq
        # print("checkbox3")

        wb.setChecked(False)

        # TODO The fact that emit is required controversial. Can one also check/uncheck without clicking?
        wb.clicked.emit()
        self.assertEqual(self.bool.value, False)

        # print("checkbox4")
        wb.setChecked(True)
        wb.clicked.emit(True)
        self.assertEqual(self.bool.value, True)

    def test_qcombo_box(self):
        wc = QtWidgets.QComboBox()

        wc.clear()
        wc.addItem("-1", -1)
        wc.addItem("-2", -2)
        wc.setCurrentIndex(1)

        # test on connection widget take lq value
        self.choices.connect_to_widget(wc)
        self.assertEqual(wc.currentData(), 1)

        ## potentially too strong of a test.
        self.assertEqual(wc.currentIndex(), 0)

        # lq to widget
        self.choices.update_value(1)
        self.assertEqual(wc.currentData(), 1)

        # widget to lq
        wc.setCurrentIndex(2)
        self.assertEqual(self.choices.value, 3)

    def test_qslider_int_non_overflow(self):
        lq = self.settings.New("int2", int, vmin=-100, vmax=100)

        wi = QtWidgets.QSlider()
        wi.setValue(3)

        # test on connection widget take lq value
        lq.connect_to_widget(wi)
        self.assertEqual(wi.value(), 0.0)

        # lq to widget
        lq.update_value(1)
        self.assertEqual(wi.value(), 1)

        # widget to lq
        wi.setValue(2)
        self.assertEqual(lq.value, 2)

    def test_qslider_int_overflow(self):
        # falls back to act like a float
        wi = QtWidgets.QSlider()
        wi.setValue(int(1000))

        lq = self.int

        # test on connection widget take lq value
        lq.connect_to_widget(wi)
        self.assertAlmostEqual(lq._transform_from_slider(wi.value()), 0.0)

        # lq to widget
        lq.update_value(int(1e11))
        self.assertAlmostEqual(
            lq._transform_from_slider(wi.value()) / 1e12, int(1e11) / 1e12
        )

        # widget to lq
        wi.setValue(lq._transform_to_slider(2e11))
        self.assertAlmostEqual(lq.value / 1e12, 2e11 / 1e12)

    def test_qslider_float(self):
        wf = QtWidgets.QSlider()
        wf.setValue(3)

        # test on connection widget take lq value
        self.float.connect_to_widget(wf)
        self.assertAlmostEqual(self.float._transform_from_slider(wf.value()), 0.0)

        # lq to widget
        self.float.update_value(1.0)
        self.assertAlmostEqual(self.float._transform_from_slider(wf.value()), 1.0)

        # widget to lq
        wf.setValue(self.float._transform_to_slider(2))
        self.assertAlmostEqual(self.float.value, 2)

    def test_min_max_slider(self):
        wf = MinMaxQSlider()
        wi = MinMaxQSlider()

        wf.update_value(3.0)
        wi.update_value(3)

        # test on connection widget take lq value
        self.float.connect_to_widget(wf)
        self.int.connect_to_widget(wi)
        self.assertEqual(wf.val, 0.0)
        self.assertEqual(wi.val, 0.0)

        # lq to widget
        self.float.update_value(1.0)
        self.int.update_value(1)
        self.assertEqual(wf.val, 1)
        self.assertEqual(wi.val, 1)

        # widget to lq
        wf.update_value(2)
        wi.update_value(2)
        self.assertEqual(self.float.value, 2)
        self.assertEqual(self.int.value, 2)

    def test_connect_to_file_widget(self):
        wdir = QtWidgets.QLineEdit()
        browseButton = QtWidgets.QPushButton("...")

        # test on connection widget take lq value
        wdir.setText("3")
        self._dir.connect_to_browse_widgets(wdir, browseButton)
        self.assertEqual(wdir.text(), "test_dir")

        # lq to widget
        self._dir.update_value("1")
        self.assertEqual(wdir.text(), "1")

        # widget to lq
        wdir.setText("2")
        wdir.editingFinished.emit()
        self.assertEqual(self._dir.value, "2")


class MyBaseApp(BaseApp):

    def __init__(self, argv, **kwargs):
        super().__init__(argv, **kwargs)

        lq = self.settings.New("test_lq", float, 33, "Hz", vmin=10, vmax=133)
        # lq = self.settings.New("test_lq", int, 33, "Hz", vmin=10, vmax=133)

        # w = MinMaxQSlider("test")
        w = QtWidgets.QSlider()
        lq.connect_to_widget(w)
        s = self.settings.New_UI(title="settings")

        self.ui = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(self.ui)
        layout.addWidget(w)
        layout.addWidget(s)
        self.ui.show()


if __name__ == "__main__":
    import sys

    app = MyBaseApp([])
    sys.exit(app.exec_())
    # unittest.main()
